#include "stdafx.h"
#include "TwainInterface.h"

#define STR_ERROR_TITLE			_T("Error")
#define STR_ERROR_DSM_FAIL		_T("Fail to load DSM.")

static const TblMessage ReturnCode[MAX_RETURNCODE] =
{
	{_T("TWRC_SUCCESS"),					TWRC_SUCCESS},
	{_T("TWRC_FAILURE"),					TWRC_FAILURE},
	{_T("TWRC_CHECKSTATUS"),				TWRC_CHECKSTATUS},
	{_T("TWRC_CANCEL"),						TWRC_CANCEL},
	{_T("TWRC_DSEVENT"),					TWRC_DSEVENT},
	{_T("TWRC_NOTDSEVENT"),					TWRC_NOTDSEVENT},
	{_T("TWRC_XFERDONE"),					TWRC_XFERDONE},
	{_T("TWRC_ENDOFLIST"),					TWRC_ENDOFLIST},
	{_T("TWRC_INFONOTSUPPORTED"),			TWRC_INFONOTSUPPORTED},
	{_T("TWRC_DATANOTAVAILABLE"),			TWRC_DATANOTAVAILABLE}
};

static const TblMessage ConditionCode[MAX_CONDITIONCODE] =
{
	{_T("TWCC_SUCCESS"),					TWCC_SUCCESS},
	{_T("TWCC_BUMMER"),						TWCC_BUMMER},
	{_T("TWCC_LOWMEMORY"),					TWCC_LOWMEMORY},
	{_T("TWCC_NODS"),						TWCC_NODS},
	{_T("TWCC_MAXCONNECTIONS"),				TWCC_MAXCONNECTIONS},
	{_T("TWCC_OPERATIONERROR"),				TWCC_OPERATIONERROR},
	{_T("TWCC_BADCAP"),						TWCC_BADCAP},
	{_T("TWCC_BADPROTOCOL"),				TWCC_BADPROTOCOL},
	{_T("TWCC_BADVALUE"),					TWCC_BADVALUE},
	{_T("TWCC_SEQERROR"),					TWCC_SEQERROR},
	{_T("TWCC_BADDEST"),					TWCC_BADDEST},
	{_T("TWCC_CAPUNSUPPORTED"),				TWCC_CAPUNSUPPORTED},
	{_T("TWCC_CAPBADOPERATION"),			TWCC_CAPBADOPERATION},
	{_T("TWCC_CAPSEQERROR"),				TWCC_CAPSEQERROR},
	{_T("TWCC_DENIED"),						TWCC_DENIED},
	{_T("TWCC_FILEEXISTS"),					TWCC_FILEEXISTS},
	{_T("TWCC_FILENOTFOUND"),				TWCC_FILENOTFOUND},
	{_T("TWCC_NOTEMPTY"),					TWCC_NOTEMPTY},
	{_T("TWCC_PAPERJAM"),					TWCC_PAPERJAM},
	{_T("TWCC_PAPERDOUBLEFEED"),			TWCC_PAPERDOUBLEFEED},
	{_T("TWCC_FILEWRITEERROR"),				TWCC_FILEWRITEERROR},
	{_T("TWCC_CHECKDEVICEONLINE"),			TWCC_CHECKDEVICEONLINE},
	{_T("TWCC_EPCS_NETWORKUNREACHABLE"),	TWCC_EPCS_NETWORKUNREACHABLE},
	{_T("TWCC_EPCS_PAPERTIMEOUT"),			TWCC_EPCS_PAPERTIMEOUT}
};

static const TblMessage DataGroup[MAX_DG] =
{
	{_T("DG_CONTROL"),						DG_CONTROL},
	{_T("DG_IMAGE"),						DG_IMAGE},
	{_T("DG_AUDIO"),						DG_AUDIO}
};

static const TblMessage DataType[MAX_DAT] =
{
	{_T("DAT_AUDIOFILEXFER"),				DAT_AUDIOFILEXFER},
	{_T("DAT_AUDIOINFO"),					DAT_AUDIOINFO},
	{_T("DAT_AUDIONATIVEXFER"),				DAT_AUDIONATIVEXFER},
	{_T("DAT_CAPABILITY"),					DAT_CAPABILITY},
	{_T("DAT_CUSTOMDSDATA"),				DAT_CUSTOMDSDATA},
	{_T("DAT_CIECOLOR"),					DAT_CIECOLOR},
	{_T("DAT_DEVICEEVENT"),					DAT_DEVICEEVENT},
	{_T("DAT_EVENT"),						DAT_EVENT},
	{_T("DAT_EXTIMAGEINFO"),				DAT_EXTIMAGEINFO},
	{_T("DAT_FILESYSTEM"),					DAT_FILESYSTEM},
	{_T("DAT_GRAYRESPONSE"),				DAT_GRAYRESPONSE},
	{_T("DAT_IDENTITY"),					DAT_IDENTITY},
	{_T("DAT_IMAGEFILEXFER"),				DAT_IMAGEFILEXFER},
	{_T("DAT_IMAGEINFO"),					DAT_IMAGEINFO},
	{_T("DAT_IMAGELAYOUT"),					DAT_IMAGELAYOUT},
	{_T("DAT_IMAGEMEMXFER"),				DAT_IMAGEMEMXFER},
	{_T("DAT_IMAGENATIVEXFER"),				DAT_IMAGENATIVEXFER},
	{_T("DAT_JPEGCOMPRESSION"),				DAT_JPEGCOMPRESSION},
	{_T("DAT_NULL"),						DAT_NULL},
	{_T("DAT_PALETTE8"),					DAT_PALETTE8},
	{_T("DAT_PARENT"),						DAT_PARENT},
	{_T("DAT_PASSTHRU"),					DAT_PASSTHRU},
	{_T("DAT_PENDINGXFERS"),				DAT_PENDINGXFERS},
	{_T("DAT_RGBRESPONSE"),					DAT_RGBRESPONSE},
	{_T("DAT_SETUPFILEXFER"),				DAT_SETUPFILEXFER},
	{_T("DAT_SETUPMEMXFER"),				DAT_SETUPMEMXFER},
	{_T("DAT_STATUS"),						DAT_STATUS},
	{_T("DAT_USERINTERFACE"),				DAT_USERINTERFACE},
	{_T("DAT_XFERGROUP"),					DAT_XFERGROUP}
};

static const TblMessage Message[MAX_MSG] =
{
	{_T("MSG_CHECKSTATUS"),					MSG_CHECKSTATUS},
	{_T("MSG_CLOSEDS"),						MSG_CLOSEDS},
	{_T("MSG_CLOSEDSM"),					MSG_CLOSEDSM},
	{_T("MSG_CLOSEDSOK"),					MSG_CLOSEDSOK},
	{_T("MSG_CLOSEDSREQ"),					MSG_CLOSEDSREQ},
	{_T("MSG_DISABLEDS"),					MSG_DISABLEDS},
	{_T("MSG_ENABLEDS"),					MSG_ENABLEDS},
	{_T("MSG_ENABLEDSUIONLY"),				MSG_ENABLEDSUIONLY},
	{_T("MSG_ENDXFER"),						MSG_ENDXFER},
	{_T("MSG_GET"),							MSG_GET},
	{_T("MSG_GETCURRENT"),					MSG_GETCURRENT},
	{_T("MSG_GETDEFAULT"),					MSG_GETDEFAULT},
	{_T("MSG_GETFIRST"),					MSG_GETFIRST},
	{_T("MSG_GETNEXT"),						MSG_GETNEXT},
	{_T("MSG_NULL"),						MSG_NULL},
	{_T("MSG_OPENDS"),						MSG_OPENDS},
	{_T("MSG_OPENDSM"),						MSG_OPENDSM},
	{_T("MSG_PROCESSEVENT"),				MSG_PROCESSEVENT},
	{_T("MSG_QUERYSUPPORT"),				MSG_QUERYSUPPORT},
	{_T("MSG_RESET"),						MSG_RESET},
	{_T("MSG_RESETALL"),					MSG_RESETALL},
	{_T("MSG_SET"),							MSG_SET},
	{_T("MSG_USERSELECT"),					MSG_USERSELECT},
	{_T("MSG_XFERREADY"),					MSG_XFERREADY},
	{_T("MSG_DEVICEEVENT"),					MSG_DEVICEEVENT},
	{_T("MSG_CHANGEDIRECTORY"),				MSG_CHANGEDIRECTORY},
	{_T("MSG_CREATEDIRECTORY"),				MSG_CREATEDIRECTORY},
	{_T("MSG_DELETE"),						MSG_DELETE},
	{_T("MSG_FORMATMEDIA"),					MSG_FORMATMEDIA},
	{_T("MSG_GETCLOSE"),					MSG_GETCLOSE},
	{_T("MSG_GETFIRSTFILE"),				MSG_GETFIRSTFILE},
	{_T("MSG_GETINFO"),						MSG_GETINFO},
	{_T("MSG_GETNEXTFILE"),					MSG_GETNEXTFILE},
	{_T("MSG_RENAME"),						MSG_RENAME},
	{_T("MSG_PASSTHRU"),					MSG_PASSTHRU}
};

//========================================================================
TwainInterface::TwainInterface()
{
	memset(&m_DSMEntry, 0, sizeof(m_DSMEntry));
	memset(&m_MyInfo, 0, sizeof(m_MyInfo));
	memset(&m_DataSource, 0, sizeof(m_DataSource));

	m_hDSM			= NULL;
	m_status		= STATE_CLOSE_DSM;

}

//========================================================================
TwainInterface::~TwainInterface()
{
	if(m_hDSM)
	{
		FreeLibrary(m_hDSM);
		m_hDSM = NULL;
	}
}

//========================================================================
void TwainInterface::fillIdentity(TW_IDENTITY& _identity)
{
	_identity.Id = 0;
	_identity.Version.MajorNum	= 1;
	_identity.Version.MinorNum	= 0;
	_identity.Version.Language	= TWLG_ENGLISH_CANADIAN;
	_identity.Version.Country	= TWCY_CANADA;
	_identity.ProtocolMajor		= TWON_PROTOCOLMAJOR;
	_identity.ProtocolMinor		= TWON_PROTOCOLMINOR;
	_identity.SupportedGroups = DF_APP2 | DG_IMAGE | DG_CONTROL;
	_tcscpy_s(_identity.Manufacturer, sizeof(_identity.Manufacturer), "EPSON");
	_tcscpy_s(_identity.ProductFamily, sizeof(_identity.ProductFamily), "EPSON Scan for Business");
	_tcscpy_s(_identity.ProductName, sizeof(_identity.ProductName), "TWAIN Sample Program");
}

//========================================================================
TW_UINT16 TwainInterface::DSM_Entry(TW_UINT32 _DG,TW_UINT16 _DAT, TW_UINT16 _MSG, TW_MEMREF _pData, BOOL bShowStatus/* = TRUE*/)
{
	TW_UINT16	twrc	= TWRC_SUCCESS;

	if(0 != m_DSMEntry.DSM_Entry)
	{
		twrc = m_DSMEntry.DSM_Entry(&m_MyInfo,
									&m_DataSource,
									_DG,
									_DAT,
									_MSG,
									_pData);

		if((TWRC_FAILURE		== twrc)	||
				(TWRC_CHECKSTATUS	== twrc))
		{
			if(bShowStatus)
			{
				GetTwainStatus(NULL, _DG, _DAT, _MSG, twrc);
			}
		}
	}
	else
	{
		twrc = TWRC_FAILURE;
	}

	return twrc;
}

//========================================================================
TW_HANDLE TwainInterface::DSM_Alloc(TW_UINT32 _size)
{
	TW_HANDLE	handle = NULL;

	if(0 != m_DSMEntry.DSM_MemAllocate)
	{
		handle = m_DSMEntry.DSM_MemAllocate(_size);
	}
	else
	{
		handle = (TW_HANDLE)::GlobalAlloc(GHND, _size);
	}

	return handle;
}

//========================================================================
void TwainInterface::DSM_Free(TW_HANDLE _handle)
{
	if(0 != m_DSMEntry.DSM_MemFree)
	{
		m_DSMEntry.DSM_MemFree(_handle);
	}
	else
	{
		::GlobalFree((HGLOBAL)_handle);
	}
}

//========================================================================
TW_MEMREF TwainInterface::DSM_Lock(TW_HANDLE _handle)
{
	TW_MEMREF	memRef = NULL;

	if(0 != m_DSMEntry.DSM_MemLock)
	{
		memRef = m_DSMEntry.DSM_MemLock(_handle);
	}
	else
	{
		memRef = (TW_MEMREF)::GlobalLock((HGLOBAL)_handle);
	}

	return memRef;
}

//========================================================================
void TwainInterface::DSM_Unlock(TW_HANDLE _handle)
{
	if(0 != m_DSMEntry.DSM_MemUnlock)
	{
		m_DSMEntry.DSM_MemUnlock(_handle);
	}
	else
	{
		GlobalUnlock((HGLOBAL)_handle);
	}
}

//========================================================================
BOOL TwainInterface::connectDSM(HWND hParent)
{
	BOOL		bSuccess	= TRUE;
	TW_UINT16	twrc		= 0;

	fillIdentity(m_MyInfo);

	if(bSuccess)
	{
		if(m_hDSM)
		{
			bSuccess = FALSE;
		}
	}

	if(bSuccess)
	{
		m_hDSM = LoadLibrary(_T("TWAINDSM.dll"));

		if(!m_hDSM)
		{
			m_hDSM = LoadLibrary(_T("TWAIN_32.DLL"));
		}

		if(!m_hDSM)
		{
			bSuccess = FALSE;
			MessageBox(hParent, STR_ERROR_DSM_FAIL, STR_ERROR_TITLE, MB_OK|MB_ICONERROR);
		}
	}

	if(bSuccess)
	{
		m_DSMEntry.DSM_Entry = (DSMENTRYPROC)GetProcAddress(m_hDSM, MAKEINTRESOURCE(1));

		if(!m_DSMEntry.DSM_Entry)
		{
			bSuccess = FALSE;
			MessageBox(hParent, STR_ERROR_DSM_FAIL, STR_ERROR_TITLE, MB_OK|MB_ICONERROR);
		}
	}

	if(bSuccess)
	{
		twrc = m_DSMEntry.DSM_Entry(&m_MyInfo,
									NULL,
									DG_CONTROL,
									DAT_PARENT,
									MSG_OPENDSM,
									(TW_MEMREF)&hParent);

		if(TWRC_SUCCESS != twrc)
		{
			GetTwainStatus(hParent, DG_CONTROL, DAT_PARENT, MSG_OPENDSM, twrc);
			bSuccess = FALSE;
		}
	}

	if(bSuccess)
	{
		if((m_MyInfo.SupportedGroups & DF_DSM2) == DF_DSM2)
		{
			m_DSMEntry.Size = sizeof(TW_ENTRYPOINT);

			twrc = m_DSMEntry.DSM_Entry(&m_MyInfo,
										NULL,
										DG_CONTROL,
										DAT_ENTRYPOINT,
										MSG_GET,
										(TW_MEMREF)&m_DSMEntry);

			if(TWRC_SUCCESS != twrc)
			{
				GetTwainStatus(hParent, DG_CONTROL, DAT_ENTRYPOINT, MSG_GET, twrc);
				bSuccess = FALSE;
			}
		}
	}

	if(bSuccess)
	{
		m_status = STATE_OPEN_DSM;
	}

	return bSuccess;
}

//========================================================================
BOOL TwainInterface::disconnectDSM(HWND hParent)
{
	BOOL		bSuccess	= TRUE;
	TW_UINT16	twrc		= 0;

	if(!m_hDSM)
		bSuccess = FALSE;
	else if(m_status > STATE_OPEN_DSM)
		bSuccess = unloadDS(hParent);

	if(bSuccess)
	{
		twrc = m_DSMEntry.DSM_Entry(&m_MyInfo,
									NULL,
									DG_CONTROL,
									DAT_PARENT,
									MSG_CLOSEDSM,
									(TW_MEMREF)&hParent);

		if(TWRC_SUCCESS != twrc)
		{
			GetTwainStatus(hParent, DG_CONTROL, DAT_PARENT, MSG_CLOSEDSM, twrc);
			bSuccess = FALSE;
		}
	}

	if(bSuccess)
	{
		FreeLibrary((HMODULE)m_hDSM);
		m_hDSM = NULL;
	}

	if(bSuccess)
	{
		m_status = STATE_CLOSE_DSM;
	}

	return bSuccess;
}

//========================================================================
BOOL TwainInterface::selectDS(HWND hParent,  pTW_IDENTITY pDsIdentity)
{
	BOOL		bSuccess	= TRUE;
	TW_UINT16	twrc		= 0;

	if(m_status < STATE_OPEN_DSM)
		bSuccess = connectDSM(hParent);
	else if(m_status >= STATE_LOAD_DS)
		bSuccess = unloadDS(hParent);

	if(bSuccess)
	{
		twrc = m_DSMEntry.DSM_Entry(&m_MyInfo,
									NULL,
									DG_CONTROL,
									DAT_IDENTITY,
									MSG_USERSELECT,
									(TW_MEMREF)&m_DataSource);

		if(TWRC_SUCCESS != twrc)
		{
			GetTwainStatus(hParent, DG_CONTROL, DAT_IDENTITY, MSG_USERSELECT, twrc);
			m_DataSource.Id = 0;
			bSuccess = FALSE;
		}
	}

	if(bSuccess)
	{
		bSuccess = loadDS(hParent);
	}

	if(bSuccess)
	{
		memcpy_s(pDsIdentity, sizeof(TW_IDENTITY), &m_DataSource, sizeof(m_DataSource));
	}

	return bSuccess;
}
//========================================================================
BOOL TwainInterface::loadDS(HWND hParent)
{
	BOOL		bSuccess	= TRUE;
	TW_UINT16	twrc		= 0;

	if(m_status < STATE_OPEN_DSM)
		bSuccess = connectDSM(hParent);
	else if(m_status > STATE_LOAD_DS)
		bSuccess = FALSE;

	if(bSuccess)
	{
		m_DataSource.Id = 0;
		twrc = m_DSMEntry.DSM_Entry(&m_MyInfo,
									NULL,
									DG_CONTROL,
									DAT_IDENTITY,
									MSG_OPENDS,
									(TW_MEMREF)&m_DataSource);

		if(TWRC_SUCCESS != twrc)
		{
			GetTwainStatus(hParent, DG_CONTROL, DAT_IDENTITY, MSG_OPENDS, twrc);
			bSuccess = FALSE;
		}
	}

	if(bSuccess)
	{
		m_status = STATE_LOAD_DS;
	}

	return bSuccess;
}

//========================================================================
BOOL TwainInterface::disableDS(HWND hParent)
{
	BOOL		bSuccess	= TRUE;
	TW_UINT16	twrc		= 0;

	if(m_status != STATE_ENABLE_DS)
		bSuccess = FALSE;

	if(bSuccess)
	{
		TW_USERINTERFACE	twUI;
		memset(&twUI, NULL, sizeof(twUI));

		twUI.ShowUI		= TWON_DONTCARE8;
		twUI.hParent	= hParent;

		twrc = m_DSMEntry.DSM_Entry(&m_MyInfo,
									&m_DataSource,
									DG_CONTROL,
									DAT_USERINTERFACE,
									MSG_DISABLEDS,
									(TW_MEMREF)&twUI);

		if(TWRC_SUCCESS != twrc)
		{
			GetTwainStatus(hParent, DG_CONTROL, DAT_USERINTERFACE, MSG_DISABLEDS, twrc);
			bSuccess = FALSE;
		}
	}

	if(bSuccess)
	{
		m_status = STATE_LOAD_DS;
	}

	return bSuccess;
}

//========================================================================
BOOL TwainInterface::unloadDS(HWND hParent)
{
	BOOL		bSuccess	= TRUE;
	TW_UINT16	twrc		= 0;

	if(m_status > STATE_LOAD_DS)
		bSuccess = disableDS(hParent);
	else if(m_status < STATE_LOAD_DS)
		bSuccess = FALSE;

	if(bSuccess)
	{
		twrc = m_DSMEntry.DSM_Entry(&m_MyInfo,
									NULL,
									DG_CONTROL,
									DAT_IDENTITY,
									MSG_CLOSEDS,
									(TW_MEMREF)&m_DataSource);

		if(TWRC_SUCCESS != twrc)
		{
			GetTwainStatus(hParent, DG_CONTROL, DAT_IDENTITY, MSG_CLOSEDS, twrc);
			bSuccess = FALSE;
		}
	}

	if(bSuccess)
	{
		m_DataSource.Id = 0;
		m_status = STATE_OPEN_DSM;
	}

	return bSuccess;
}

//========================================================================
void TwainInterface::GetTwainStatus(HWND hParent, TW_UINT32 _DG, TW_UINT16 _DAT, TW_UINT16 _MSG, TW_UINT16 _twrc)
{
	TW_UINT16			twrc		= TWRC_SUCCESS;
	TW_STATUS			status;

	status.ConditionCode	= 0;
	status.Reserved			= 0;

	if(m_DSMEntry.DSM_Entry)
	{
		twrc = m_DSMEntry.DSM_Entry(&m_MyInfo,
									&m_DataSource,
									DG_CONTROL,
									DAT_STATUS,
									MSG_GET,
									&status);

		if((TWRC_SUCCESS == twrc) && (TWCC_SUCCESS != status.ConditionCode))
		{
			BOOL		bFind	= FALSE;
			TW_STR255	str;
			TW_STR255	strTmp;
			ZeroMemory(str, sizeof(str));
			ZeroMemory(strTmp, sizeof(strTmp));

			for(TW_UINT32 i = 0; i < MAX_RETURNCODE; i++)
			{
				if(ReturnCode[i].ItemId == _twrc)
				{
					_stprintf_s(strTmp, sizeof(strTmp), STR_FORMAT_RETURNCODE, ReturnCode[i].tcsItemName);
					_tcscat_s(str, sizeof(str), strTmp);
					bFind = TRUE;
					break;
				}
			}

			if(!bFind)
			{
				_stprintf_s(str, sizeof(str), STR_FORMAT_RETURNCODE_UNKNOWN, _twrc);
				_tcscat_s(str, sizeof(str), strTmp);
			}
			_tcscat_s(str, sizeof(str), _T("\n"));

			bFind = FALSE;
			for(TW_UINT32 i = 0; i < MAX_CONDITIONCODE; i++)
			{
				if(ConditionCode[i].ItemId == status.ConditionCode)
				{
					_stprintf_s(strTmp, sizeof(strTmp), STR_FORMAT_CONDITIONCODE, ConditionCode[i].tcsItemName);
					_tcscat_s(str, sizeof(str), strTmp);
					bFind = TRUE;
					break;
				}
			}

			if(!bFind)
			{
				_stprintf_s(str, sizeof(str), STR_FORMAT_CONDITIONCODE_UNKNOWN, status.ConditionCode);
				_tcscat_s(str, sizeof(str), strTmp);
			}
			_tcscat_s(str, sizeof(str), _T("\n"));
			_tcscat_s(str, sizeof(str), _T("\n"));

			bFind = FALSE;
			for(TW_UINT32 i = 0; i < MAX_DG; i++)
			{
				if(DataGroup[i].ItemId == _DG)
				{
					_stprintf_s(strTmp, sizeof(strTmp), STR_FORMAT_DG, DataGroup[i].tcsItemName);
					_tcscat_s(str, sizeof(str), strTmp);
					bFind = TRUE;
					break;
				}
			}

			if(!bFind)
			{
				_stprintf_s(strTmp, sizeof(strTmp), STR_FORMAT_DG_UNKNOWN, _DG);
				_tcscat_s(str, sizeof(str), strTmp);
			}
			_tcscat_s(str, sizeof(str), _T("\n"));

			bFind = FALSE;
			for(TW_UINT32 i = 0; i < MAX_DAT; i++)
			{
				if(DataType[i].ItemId == _DAT)
				{
					_stprintf_s(strTmp, sizeof(strTmp), STR_FORMAT_DAT, DataType[i].tcsItemName);
					_tcscat_s(str, sizeof(str), strTmp);
					bFind = TRUE;
					break;
				}
			}

			if(!bFind)
			{
				_stprintf_s(strTmp, sizeof(strTmp), STR_FORMAT_DAT, _DAT);
				_tcscat_s(str, sizeof(str), strTmp);
			}
			_tcscat_s(str, sizeof(str), _T("\n"));

			bFind = FALSE;
			for(TW_UINT32 i = 0; i < MAX_MSG; i++)
			{
				if(Message[i].ItemId == _MSG)
				{
					_stprintf_s(strTmp, sizeof(strTmp), STR_FORMAT_MSG, Message[i].tcsItemName);
					_tcscat_s(str, sizeof(str), strTmp);
					bFind = TRUE;
					break;
				}
			}

			if(!bFind)
			{
				_stprintf_s(strTmp, sizeof(strTmp), STR_FORMAT_MSG_UNKNOWN, _MSG);
				_tcscat_s(str, sizeof(str), strTmp);
			}
			_tcscat_s(str, sizeof(str), _T("\n"));

			MessageBox(hParent, str, STR_ERROR_TITLE, MB_OK|MB_ICONERROR);
		}
	}
}

BOOL TwainInterface::Acquire(HWND hParent, TW_BOOL bShowUI)
{
	BOOL		bSuccess	= TRUE;
	TW_UINT16	twrc		= 0;

	if(m_status < STATE_LOAD_DS)
		bSuccess = loadDS(hParent);
	else if(m_status > STATE_ENABLE_DS)
		bSuccess = FALSE;
	else
	{
		// UI Suppress Mode
		if(m_status == STATE_ENABLE_DS && bShowUI == FALSE)
			bSuccess = disableDS(hParent);
	}

	if(bSuccess)
	{
		TW_USERINTERFACE	twUI	= {0};

		twUI.ShowUI		= bShowUI;
		twUI.hParent	= hParent;

		twrc = m_DSMEntry.DSM_Entry(&m_MyInfo,
									&m_DataSource,
									DG_CONTROL,
									DAT_USERINTERFACE,
									MSG_ENABLEDS,
									(TW_MEMREF)&twUI);

		if(TWRC_SUCCESS != twrc)
		{
			GetTwainStatus(hParent, DG_CONTROL, DAT_USERINTERFACE, MSG_ENABLEDS, twrc);
			bSuccess = FALSE;
		}
	}

	if(bSuccess)
	{
		m_status = STATE_ENABLE_DS;
	}

	return bSuccess;
}

//========================================================================
STATECODE TwainInterface::GetStateCode(void)
{
	return	m_status;
}